% Author: G�nter, Andi
% Date: 23.09.2004

/**
  * max_cycles(+Set, ?DepType, ?MaxSortedCycles)
  *
  * arg1 is a list (set) of nodes in the examination graph
  * arg2 is the type of connection between the nodes
  * arg3 will be bound to lists of all nodes for each biggest strongly
  * connected part of the graph
*/

max_cycles(Set,DepType, MaxSortedCycles) :-
    cycles(Set,DepType, AllSortedCycles),
    setof( (Len,Cycle),
           (member(Cycle,AllSortedCycles),  length(Cycle, Len)),
           CyclesWithLength
           ),
      % At this point CyclesWithLength is a list of pairs
      % (Len,Cycle) that is sorted by the value of Len:
    findall(SameLengthSubList,
            same_length_sublist( CyclesWithLength, SameLengthSubList),
            ListOfSameLengthCycles
            ),
    subset_free( ListOfSameLengthCycles,SubsetFree),
    MaxSortedCycles = SubsetFree.

%test(max_cycles(Set, DepType, MaxSortedCycles)) :-
 %     testSet(Set),
  %    DepType =  packageDep,
   %   max_cycles(Set, DepType, MaxSortedCycles).
      %,
      %MaxSortedCycles = [[p4, p5], [p1, p2, p3, p6]].


/**
  * cycles(+Set, ?DepType, ?AllSortedCycles)
  *
  * arg1 is a list (set) of nodes in the examination graph
  * arg2 is the type of connection between the nodes
  * arg3 will be bound to lists of nodes of all cycles found in the graph.
  *      if subcycles are found, they will trigger also!
  */

cycles(Set,DepType, AllSortedCycles) :-
    setof( CycleSortedSet, cycle(Set, CycleSortedSet,DepType), AllSortedCycles).


/**
  * cycle(+Set, ?AllSortedCycles, ?DepType)
  *
  * arg1 is a list (set) of nodes in the examination graph
  * arg2 will be bound to lists of nodes of all cycles found in the graph.
  *    if subcycles are found, they will trigger also!
  *    same cycles may trigger more often than once
  * arg3 is the type of connection between the nodes
  */

cycle(Set, CycleSortedSet,DepType) :-
    path(Set, First, Last, [_|Path1], DepType),
    path(Set, Last, First, [_|Path2], DepType),
    collect_(Path1, Path2, CycleSortedSet).

/**
  * path(?Set, ?First, ?Last, ?Path, ?Type)
  *
  * is true if Path is a path of length Length from the
  * First vertex to the Last vertex in the graph defined by e/3
  */

path(Set, First, Last, Path, Type):-
    path__(Set, First, Last, [], Path, Type).

/**
  * path_1(+Set, ?Last, ?Last, +Path, ?LastPath, +Type)
  */
path__(Set, Last, Last, Path, [Last|Path], _) :-
    nonvar(Last),
    check_in_set_(Last, Set).

path__(Set, First, Last, Path0, Path, DepType):-
    dependency_db(NextToLast, Last, DepType, _), % <-- asserted graph elem
    not(member(NextToLast, Path0)),
    check_in_set_(Last, Set),
    check_in_set_(NextToLast, Set),
    path__(Set, First, NextToLast, [Last|Path0], Path, DepType).

check_in_set_(_, Set) :-
    var(Set), !.
check_in_set_(X, Set) :-
    member(X, Set).


collect_(Path1, Path2, CycleSortedSet) :-
    append(Path1, Path2, CycleUnsorted),
    sort(CycleUnsorted, CycleSorted),
    list_to_set(CycleSorted, CycleSortedSet).

  /**
   *  subset_free( +ListOfSameLengthCycles, ?SubsetFree)
   *
   *  arg1 is a list of lists where each sublist contains itself duplicate
   *  free lists of same length, e.g.
   *      arg1 = [[[p4, p5], [p2, p3]], [[p1, p2, p3]]].
   *  The sublists of arg1 are sorted in ascending order by the length
   *  of their own sublists. These assumptions are helpful to speed up
   *  the many comparisons that we must do in this predicate. Lists of
   *  same length do not need to be compared for being subsets of each other.
   *
   *  arg2 is a list of those inner (second nesting level) lists that
   *  are not subset of any other list, e.g.
   *      arg2 = [ [p4, p5], [p1, p2, p3] ].
   */

subset_free(ListOfSameLengthCycles, SubsetFree) :-
    subset_free_(ListOfSameLengthCycles,[],SubsetFree).

subset_free_([],Res,Res).

subset_free_( [LengthN | Longer], Nminus1, Res) :-
   no_subset(LengthN, Longer, NoSubsetN),
   append(Nminus1,NoSubsetN,NoSubset1ToN),
   subset_free_( Longer, NoSubset1ToN, Res).

/**
  * no_subset(+LengthN, +Longer, ?NoSubsetN)
  *
  * arg1 is a list of lists of same length
  * e.g. arg1 = [[p4, p5], [p2, p3]]
  * arg2 is a list of lists where each sublist contains itself duplicate
  * free lists of same length. All lists in arg2 are longer than those in arg1
  * e.g. arg2 = [[[p4, p5, p3], [p2, p3, p4]], [[p1, p2, p3, p5]]].
  * arg3 is a list containing only those elements of arg1 that are not subsets
  * of any element in arg2
  */

no_subset([], _, []).

no_subset([FirstList|Rest], Longer, NoSubsetN) :-
    no_subset_in_subsublist(FirstList,Longer),
    !,
    NoSubsetN = [FirstList| NoSubsetRest] ,
    no_subset(Rest,Longer,NoSubsetRest).

no_subset([_|Rest], Longer, NoSubsetN) :-
    no_subset(Rest,Longer,NoSubsetN).


/**
  * no_subset_in_subsublist(+FirstList,+Longer)
  *
  * arg1 is a list
  * arg2 is a list of lists where each sublist contains itself duplicate
  * free lists of same length. All lists in arg2 are longer than those in arg1
  * e.g. arg2 = [[[p4, p5, p3], [p2, p3, p4]], [[p1, p2, p3, p5]]].
  * This predicate will be true, if arg1 is a no subset of any of the inner
  * sublists of arg2
  * e.g. arg1 = [1,2,3] will succeed but arg1 = [p2,p4] or arg1 = [p3,p5]
  * will fail.
  */

no_subset_in_subsublist(_,[]).

no_subset_in_subsublist(List,[First|Rest]) :-
    no_subset_in_sublist(List,First),
    !,
    no_subset_in_subsublist(List,Rest).
    
/**
  * no_subset_in_sublist(+FirstList,+Longer)
  *
  * arg1 is a list
  * arg2 is a list containing itself duplicate
  * free lists of same length.
  * e.g. arg2 = [[p4, p5, p3], [p2, p3, p4]].
  * This predicate will be true, if arg1 is a no subset of any of the
  * sublists of arg2
  * e.g. arg1 = [1,2,3] will succeed but arg1 = [p2,p4] or arg1 = [p3,p5]
  * will fail.
  */

no_subset_in_sublist(_,[]).

no_subset_in_sublist(List,[First|Rest]) :-
    not(subset(List,First)),
    no_subset_in_sublist(List,Rest).

 /**
  * same_length_sublist(+In, ?Sub)
  *
  * arg1 is a list of pairs of the form (L,X), sorted on the values of L.
  * arg 2 is a maximal sublist of X elements with same L value.
  */


same_length_sublist(In, Sub) :-
   same_length_sublist(In,[], Sub).
   
same_length_sublist([], _, []) :- !.            % empty list:
                                                % --> has empty sublist

same_length_sublist([(_,X)], Prev, [X|Prev]) :- !.  % singleton list:
                                                % --> add its X elem to remembered
                                                
same_length_sublist([A,B|C], Prev, D) :-        % first two elements have same length:
  A = (Len,X),                                  % --> remember first element and
  B = (Len,_),                                  %     scan rest of list
  !,
  same_length_sublist([B|C], [X|Prev], D).

same_length_sublist([(_,X),_|_], Prev, [X|Prev]) .  % first two elements are different:
                                                % --> add its first X elem to remembered

same_length_sublist([_|C], _, Res) :-           % first two elements are different:
   same_length_sublist(C, [], Res).             % --> find results in rest of list

